@@ -54,7 +54,7 @@ def _verify_dict_keys(expected_keys, target_dict, strict=True):
     predicate = expected_keys.__eq__ if strict else expected_keys.issubset
 
     if not predicate(provided_keys):
-        msg = (_("Validation of dictionary's keys failed."
+        msg = (_("Validation of dictionary's keys failed. "
                  "Expected keys: %(expected_keys)s "
                  "Provided keys: %(provided_keys)s") %
                {'expected_keys': expected_keys,
@@ -153,19 +153,16 @@ def _validate_no_whitespace(data):
 
 
 def _validate_mac_address(data, valid_values=None):
-    valid_mac = False
     try:
         valid_mac = netaddr.valid_mac(_validate_no_whitespace(data))
-    except Exception:
-        pass
-    finally:
-        # TODO(arosen): The code in this file should be refactored
-        # so it catches the correct exceptions. _validate_no_whitespace
-        # raises AttributeError if data is None.
         if valid_mac is False:
             msg = _("'%s' is not a valid MAC address") % data
             LOG.debug(msg)
             return msg
+    except AttributeError as ex:
+        msg = _("MAC address must be string: %s") % ex
+        LOG.exception(msg)
+        return msg
 
 
 def _validate_mac_address_or_none(data, valid_values=None):
@@ -449,6 +446,16 @@ def _validate_dict_or_nodata(data, key_specs=None):
         return _validate_dict(data, key_specs)
 
 
+def _validate_dict_not_empty(data, key_specs=None):
+    if not isinstance(data, dict) or len(data) == 0:
+        msg = _("'%s' is not a valid dictionary or it is an"
+                " empty dictionary") % data
+        LOG.debug(msg)
+        return msg
+    else:
+        return _validate_dict(data, key_specs)
+
+
 def _validate_non_negative(data, valid_values=None):
     try:
         data = int(data)
@@ -553,6 +560,7 @@ validators = {'type:dict': _validate_dict,
               'type:dict_or_none': _validate_dict_or_none,
               'type:dict_or_empty': _validate_dict_or_empty,
               'type:dict_or_nodata': _validate_dict_or_nodata,
+              'type:dict_not_empty': _validate_dict_not_empty,
               'type:fixed_ips': _validate_fixed_ips,
               'type:hostroutes': _validate_hostroutes,
               'type:ip_address': _validate_ip_address,
