@@ -14,7 +14,8 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-
+import ast
+import copy
 import inspect
 import threading
 import time
@@ -24,11 +25,11 @@ from oslo_log import log as logging
 from oslo_serialization import jsonutils
 from oslo_utils import timeutils
 
-
 from tacker.common import driver_manager
+from tacker.common import exceptions
 from tacker import context as t_context
-from tacker.db.common_services import common_services_db_plugin
 from tacker.plugins.common import constants
+from tacker.vnfm import utils as vnfm_utils
 
 LOG = logging.getLogger(__name__)
 CONF = cfg.CONF
@@ -43,17 +44,8 @@ CONF.register_opts(OPTS, group='monitor')
 def config_opts():
     return [('monitor', OPTS),
             ('tacker', VNFMonitor.OPTS),
-            ('tacker', VNFAlarmMonitor.OPTS), ]
-
-
-def _log_monitor_events(context, vnf_dict, evt_details):
-    _cos_db_plg = common_services_db_plugin.CommonServicesPluginDb()
-    _cos_db_plg.create_event(context, res_id=vnf_dict['id'],
-                             res_type=constants.RES_TYPE_VNF,
-                             res_state=vnf_dict['status'],
-                             evt_type=constants.RES_EVT_MONITOR,
-                             tstamp=timeutils.utcnow(),
-                             details=evt_details)
+            ('tacker', VNFAlarmMonitor.OPTS),
+            ('tacker', VNFAppMonitor.OPTS)]
 
 
 class VNFMonitor(object):
@@ -95,19 +87,27 @@ class VNFMonitor(object):
             time.sleep(self._status_check_intvl)
 
             with self._lock:
-                for hosting_vnf in self._hosting_vnfs.values():
-                    if hosting_vnf.get('dead', False):
-                        LOG.debug('monitor skips dead vnf %s', hosting_vnf)
+                for hosting_vnf in VNFMonitor._hosting_vnfs.values():
+                    if hosting_vnf.get('dead', False) or (
+                            hosting_vnf['vnf']['status'] ==
+                            constants.PENDING_HEAL):
+                        LOG.debug(
+                            'monitor skips for DEAD/PENDING_HEAL vnf %s',
+                            hosting_vnf)
                         continue
-
-                    self.run_monitor(hosting_vnf)
+                    try:
+                        self.run_monitor(hosting_vnf)
+                    except Exception as ex:
+                        LOG.exception("Unknown exception: Monitoring failed "
+                                      "for VNF '%s' due to '%s' ",
+                                      hosting_vnf['id'], ex)
 
     @staticmethod
     def to_hosting_vnf(vnf_dict, action_cb):
         return {
             'id': vnf_dict['id'],
-            'management_ip_addresses': jsonutils.loads(
-                vnf_dict['mgmt_url']),
+            'mgmt_ip_addresses': jsonutils.loads(
+                vnf_dict['mgmt_ip_address']),
             'action_cb': action_cb,
             'vnf': vnf_dict,
             'monitoring_policy': jsonutils.loads(
@@ -117,36 +117,54 @@ class VNFMonitor(object):
     def add_hosting_vnf(self, new_vnf):
         LOG.debug('Adding host %(id)s, Mgmt IP %(ips)s',
                   {'id': new_vnf['id'],
-                   'ips': new_vnf['management_ip_addresses']})
+                   'ips': new_vnf['mgmt_ip_addresses']})
         new_vnf['boot_at'] = timeutils.utcnow()
         with self._lock:
-            self._hosting_vnfs[new_vnf['id']] = new_vnf
+            VNFMonitor._hosting_vnfs[new_vnf['id']] = new_vnf
 
         attrib_dict = new_vnf['vnf']['attributes']
         mon_policy_dict = attrib_dict['monitoring_policy']
         evt_details = (("VNF added for monitoring. "
                         "mon_policy_dict = %s,") % (mon_policy_dict))
-        _log_monitor_events(t_context.get_admin_context(), new_vnf['vnf'],
-                            evt_details)
+        vnfm_utils.log_events(t_context.get_admin_context(),
+                              new_vnf['vnf'],
+                              constants.RES_EVT_MONITOR, evt_details)
 
     def delete_hosting_vnf(self, vnf_id):
         LOG.debug('deleting vnf_id %(vnf_id)s', {'vnf_id': vnf_id})
         with self._lock:
-            hosting_vnf = self._hosting_vnfs.pop(vnf_id, None)
+            hosting_vnf = VNFMonitor._hosting_vnfs.pop(vnf_id, None)
             if hosting_vnf:
                 LOG.debug('deleting vnf_id %(vnf_id)s, Mgmt IP %(ips)s',
                           {'vnf_id': vnf_id,
-                           'ips': hosting_vnf['management_ip_addresses']})
+                           'ips': hosting_vnf['mgmt_ip_addresses']})
+
+    def update_hosting_vnf(self, updated_vnf_dict, evt_details=None):
+        with self._lock:
+            vnf_to_update = VNFMonitor._hosting_vnfs.get(
+                updated_vnf_dict.get('id'))
+            if vnf_to_update:
+                updated_vnf = copy.deepcopy(updated_vnf_dict)
+                vnf_to_update['vnf'] = updated_vnf
+                vnf_to_update['mgmt_ip_addresses'] = jsonutils.loads(
+                    updated_vnf_dict['mgmt_ip_address'])
+
+                if evt_details is not None:
+                    vnfm_utils.log_events(t_context.get_admin_context(),
+                                          vnf_to_update['vnf'],
+                                          constants.RES_EVT_HEAL,
+                                          evt_details=evt_details)
 
     def run_monitor(self, hosting_vnf):
-        mgmt_ips = hosting_vnf['management_ip_addresses']
+        mgmt_ips = hosting_vnf['mgmt_ip_addresses']
         vdupolicies = hosting_vnf['monitoring_policy']['vdus']
 
         vnf_delay = hosting_vnf['monitoring_policy'].get(
             'monitoring_delay', self.boot_wait)
 
         for vdu in vdupolicies.keys():
-            if hosting_vnf.get('dead'):
+            if hosting_vnf.get('dead') or (
+                    hosting_vnf['vnf']['status']) == constants.PENDING_HEAL:
                 return
 
             policy = vdupolicies[vdu]
@@ -161,8 +179,7 @@ class VNFMonitor(object):
                         continue
 
                 actions = policy[driver].get('actions', {})
-                if 'mgmt_ip' not in params:
-                    params['mgmt_ip'] = mgmt_ips[vdu]
+                params['mgmt_ip'] = mgmt_ips[vdu]
 
                 driver_return = self.monitor_call(driver,
                                                   hosting_vnf['vnf'],
@@ -172,10 +189,10 @@ class VNFMonitor(object):
 
                 if driver_return in actions:
                     action = actions[driver_return]
-                    hosting_vnf['action_cb'](action)
+                    hosting_vnf['action_cb'](action, vdu_name=vdu)
 
     def mark_dead(self, vnf_id):
-        self._hosting_vnfs[vnf_id]['dead'] = True
+        VNFMonitor._hosting_vnfs[vnf_id]['dead'] = True
 
     def _invoke(self, driver, **kwargs):
         method = inspect.stack()[1][3]
@@ -195,6 +212,49 @@ class VNFMonitor(object):
                             vnf=vnf_dict, kwargs=kwargs)
 
 
+class VNFAppMonitor(object):
+    """VNF App monitor"""
+    OPTS = [
+        cfg.ListOpt(
+            'app_monitor_driver', default=['zabbix'],
+            help=_('App monitoring driver to communicate with '
+                   'Hosting VNF/logical service '
+                   'instance tacker plugin will use')),
+    ]
+    cfg.CONF.register_opts(OPTS, 'tacker')
+
+    def __init__(self):
+        self._application_monitor_manager = driver_manager.DriverManager(
+            'tacker.tacker.app_monitor.drivers',
+            cfg.CONF.tacker.app_monitor_driver)
+
+    def _create_app_monitoring_dict(self, dev_attrs, mgmt_ip_address):
+        app_policy = 'app_monitoring_policy'
+        appmonitoring_dict = ast.literal_eval(dev_attrs[app_policy])
+        vdulist = appmonitoring_dict['vdus'].keys()
+
+        for vduname in vdulist:
+            temp = ast.literal_eval(mgmt_ip_address)
+            appmonitoring_dict['vdus'][vduname]['mgmt_ip'] = temp[vduname]
+        return appmonitoring_dict
+
+    def create_app_dict(self, context, vnf_dict):
+        dev_attrs = vnf_dict['attributes']
+        mgmt_ip_address = vnf_dict['mgmt_ip_address']
+        return self._create_app_monitoring_dict(dev_attrs, mgmt_ip_address)
+
+    def _invoke(self, driver, **kwargs):
+        method = inspect.stack()[1][3]
+        return self._application_monitor_manager.\
+            invoke(driver, method, **kwargs)
+
+    def add_to_appmonitor(self, applicationvnfdict, vnf_dict):
+        vdunode = applicationvnfdict['vdus'].keys()
+        driver = applicationvnfdict['vdus'][vdunode[0]]['name']
+        kwargs = applicationvnfdict
+        return self._invoke(driver, vnf=vnf_dict, kwargs=kwargs)
+
+
 class VNFAlarmMonitor(object):
     """VNF Alarm monitor"""
     OPTS = [
@@ -223,9 +283,9 @@ class VNFAlarmMonitor(object):
             # TODO(Tung Doan) trigger_dict.get('actions') needs to be used
             policy_action = trigger_dict.get('action')
             if len(policy_action) == 0:
-                _log_monitor_events(t_context.get_admin_context(),
-                                    vnf,
-                                    "Alarm not set: policy action missing")
+                vnfm_utils.log_events(t_context.get_admin_context(), vnf,
+                                      constants.RES_EVT_MONITOR,
+                                      "Alarm not set: policy action missing")
                 return
             # Other backend policies with the construct (policy, action)
             # ex: (SP1, in), (SP1, out)
@@ -256,13 +316,12 @@ class VNFAlarmMonitor(object):
             alarm_url[trigger_name] =\
                 self.call_alarm_url(driver, vnf, params)
             details = "Alarm URL set successfully: %s" % alarm_url
-            _log_monitor_events(t_context.get_admin_context(),
-                                vnf,
-                                details)
+            vnfm_utils.log_events(t_context.get_admin_context(), vnf,
+                                  constants.RES_EVT_MONITOR, details)
         return alarm_url
 
     def process_alarm_for_vnf(self, vnf, trigger):
-        '''call in plugin'''
+        """call in plugin"""
         params = trigger['params']
         mon_prop = trigger['trigger']
         alarm_dict = dict()
@@ -284,3 +343,87 @@ class VNFAlarmMonitor(object):
     def process_alarm(self, driver, vnf_dict, kwargs):
         return self._invoke(driver,
                             vnf=vnf_dict, kwargs=kwargs)
+
+
+class VNFReservationAlarmMonitor(VNFAlarmMonitor):
+    """VNF Reservation Alarm monitor"""
+
+    def update_vnf_with_reservation(self, plugin, context, vnf, policy_dict):
+
+        alarm_url = dict()
+
+        def create_alarm_action(action, action_list, scaling_type):
+            params = dict()
+            params['vnf_id'] = vnf['id']
+            params['mon_policy_name'] = action
+            driver = 'ceilometer'
+
+            def _refactor_backend_policy(bk_policy_name, bk_action_name):
+                policy = '%(policy_name)s%(action_name)s' % {
+                    'policy_name': bk_policy_name,
+                    'action_name': bk_action_name}
+                return policy
+
+            for index, policy_action_name in enumerate(action_list):
+                filters = {'name': policy_action_name}
+                bkend_policies = \
+                    plugin.get_vnf_policies(context, vnf['id'], filters)
+                if bkend_policies:
+                    if constants.POLICY_SCALING in str(bkend_policies[0]):
+                        action_list[index] = _refactor_backend_policy(
+                            policy_action_name, scaling_type)
+
+                # Support multiple action. Ex: respawn % notify
+                action_name = '%'.join(action_list)
+                params['mon_policy_action'] = action_name
+                alarm_url[action] = \
+                    self.call_alarm_url(driver, vnf, params)
+                details = "Alarm URL set successfully: %s" % alarm_url
+                vnfm_utils.log_events(t_context.get_admin_context(), vnf,
+                                      constants.RES_EVT_MONITOR,
+                                      details)
+
+        before_end_action = policy_dict['reservation']['before_end_actions']
+        end_action = policy_dict['reservation']['end_actions']
+        start_action = policy_dict['reservation']['start_actions']
+
+        scaling_policies = \
+            plugin.get_vnf_policies(
+                context, vnf['id'], filters={
+                    'type': constants.POLICY_SCALING})
+
+        if len(scaling_policies) == 0:
+            raise exceptions.VnfPolicyNotFound(
+                policy=constants.POLICY_SCALING, vnf_id=vnf['id'])
+
+        for scaling_policy in scaling_policies:
+            # validating start_action for scale-out policy action
+            if scaling_policy['name'] not in start_action:
+                raise exceptions.Invalid(
+                    'Not a valid template: start_action must contain'
+                    ' %s as scaling-out action' % scaling_policy['name'])
+
+            # validating before_end and end_actions for scale-in policy action
+            if scaling_policy['name'] not in before_end_action:
+                if scaling_policy['name'] not in end_action:
+                    raise exceptions.Invalid(
+                        'Not a valid template:'
+                        ' before_end_action or end_action'
+                        ' should contain scaling policy: %s'
+                        % scaling_policy['name'])
+
+        for action in constants.RESERVATION_POLICY_ACTIONS:
+            scaling_type = "-out" if action == 'start_actions' else "-in"
+            create_alarm_action(action, policy_dict[
+                'reservation'][action], scaling_type)
+
+        return alarm_url
+
+    def process_alarm_for_vnf(self, vnf, trigger):
+        """call in plugin"""
+        params = trigger['params']
+        alarm_dict = dict()
+        alarm_dict['alarm_id'] = params['data'].get('alarm_id')
+        alarm_dict['status'] = params['data'].get('current')
+        driver = 'ceilometer'
+        return self.process_alarm(driver, vnf, alarm_dict)
