@@ -29,6 +29,7 @@ from tacker.api.v1 import attributes
 from tacker.common import driver_manager
 from tacker.common import exceptions
 from tacker.common import utils
+from tacker import context as t_context
 from tacker.db.vnfm import vnfm_db
 from tacker.extensions import vnfm
 from tacker.plugins.common import constants
@@ -98,7 +99,7 @@ class VNFMMgmtMixin(object):
         return self._invoke(
             vnf_dict, plugin=self, context=context, vnf=vnf_dict)
 
-    def mgmt_url(self, context, vnf_dict):
+    def mgmt_ip_address(self, context, vnf_dict):
         return self._invoke(
             vnf_dict, plugin=self, context=context, vnf=vnf_dict)
 
@@ -123,7 +124,7 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
     OPTS_POLICY_ACTION = [
         cfg.ListOpt(
             'policy_action', default=['autoscaling', 'respawn',
-                                      'log', 'log_and_kill'],
+                                      'vdu_autoheal', 'log', 'log_and_kill'],
             help=_('Hosting vnf drivers tacker plugin will use')),
     ]
     cfg.CONF.register_opts(OPTS_POLICY_ACTION, 'tacker')
@@ -143,7 +144,18 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
             cfg.CONF.tacker.policy_action)
         self._vnf_monitor = monitor.VNFMonitor(self.boot_wait)
         self._vnf_alarm_monitor = monitor.VNFAlarmMonitor()
+        self._vnf_reservation_monitor = monitor.VNFReservationAlarmMonitor()
         self._vnf_app_monitor = monitor.VNFAppMonitor()
+        self._init_monitoring()
+
+    def _init_monitoring(self):
+        context = t_context.get_admin_context()
+        vnfs = self.get_vnfs(context)
+        for vnf in vnfs:
+            # Add tenant_id in context object as it is required
+            # to get VIM in monitoring.
+            context.tenant_id = vnf['tenant_id']
+            self.add_vnf_to_monitor(context, vnf)
 
     def spawn_n(self, function, *args, **kwargs):
         self._pool.spawn_n(function, *args, **kwargs)
@@ -157,8 +169,8 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
             # key value string pairs in vnf attributes table
             vnfd_data['attributes']['vnfd'] = yaml.safe_dump(
                 template)
-        elif isinstance(template, str):
-            self._report_deprecated_yaml_str()
+        else:
+            raise vnfm.InvalidAPIAttributeType(atype=type(template))
         if "tosca_definitions_version" not in template:
             raise exceptions.Invalid('Not a valid template: '
                                      'tosca_definitions_version is missing.')
@@ -226,13 +238,13 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
 
     def add_vnf_to_monitor(self, context, vnf_dict):
         dev_attrs = vnf_dict['attributes']
-        mgmt_url = vnf_dict['mgmt_url']
-        if 'monitoring_policy' in dev_attrs and mgmt_url:
-            def action_cb(action):
+        mgmt_ip_address = vnf_dict['mgmt_ip_address']
+        if 'monitoring_policy' in dev_attrs and mgmt_ip_address:
+            def action_cb(action, **kwargs):
                 LOG.debug('policy action: %s', action)
                 self._vnf_action.invoke(
                     action, 'execute_action', plugin=self, context=context,
-                    vnf_dict=hosting_vnf['vnf'], args={})
+                    vnf_dict=hosting_vnf['vnf'], args=kwargs)
 
             hosting_vnf = self._vnf_monitor.to_hosting_vnf(
                 vnf_dict, action_cb)
@@ -242,17 +254,23 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
     def add_alarm_url_to_vnf(self, context, vnf_dict):
         vnfd_yaml = vnf_dict['vnfd']['attributes'].get('vnfd', '')
         vnfd_dict = yaml.safe_load(vnfd_yaml)
-        if vnfd_dict and vnfd_dict.get('tosca_definitions_version'):
-            polices = vnfd_dict['topology_template'].get('policies', [])
-            for policy_dict in polices:
-                name, policy = list(policy_dict.items())[0]
-                if policy['type'] in constants.POLICY_ALARMING:
-                    alarm_url =\
-                        self._vnf_alarm_monitor.update_vnf_with_alarm(
-                            self, context, vnf_dict, policy)
-                    vnf_dict['attributes']['alarming_policy'] = vnf_dict['id']
-                    vnf_dict['attributes'].update(alarm_url)
-                    break
+        if not (vnfd_dict and vnfd_dict.get('tosca_definitions_version')):
+            return
+        polices = vnfd_dict['topology_template'].get('policies', [])
+        for policy_dict in polices:
+            name, policy = list(policy_dict.items())[0]
+            if policy['type'] in constants.POLICY_ALARMING:
+                alarm_url =\
+                    self._vnf_alarm_monitor.update_vnf_with_alarm(
+                        self, context, vnf_dict, policy)
+                vnf_dict['attributes']['alarming_policy'] = vnf_dict['id']
+                vnf_dict['attributes'].update(alarm_url)
+            elif policy['type'] in constants.POLICY_RESERVATION:
+                alarm_url = \
+                    self._vnf_reservation_monitor.update_vnf_with_reservation(
+                        self, context, vnf_dict, policy)
+                vnf_dict['attributes']['reservation_policy'] = vnf_dict['id']
+                vnf_dict['attributes'].update(alarm_url)
 
     def add_vnf_to_appmonitor(self, context, vnf_dict):
         appmonitor = self._vnf_app_monitor.create_app_dict(context, vnf_dict)
@@ -262,6 +280,12 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
         config = vnf_dict['attributes'].get('config')
         if not config:
             return
+        if isinstance(config, str):
+            # TODO(dkushwaha) remove this load once db supports storing
+            # json format of yaml files in a separate column instead of
+            #  key value string pairs in vnf attributes table.
+            config = yaml.safe_load(config)
+
         eventlet.sleep(self.boot_wait)      # wait for vm to be ready
         vnf_id = vnf_dict['id']
         update = {
@@ -294,20 +318,20 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
                                              six.text_type(e))
 
         if instance_id is None or create_failed:
-            mgmt_url = None
+            mgmt_ip_address = None
         else:
-            # mgmt_url = self.mgmt_url(context, vnf_dict)
+            # mgmt_ip_address = self.mgmt_ip_address(context, vnf_dict)
             # FIXME(yamahata):
-            mgmt_url = vnf_dict['mgmt_url']
+            mgmt_ip_address = vnf_dict['mgmt_ip_address']
 
         self._create_vnf_post(
-            context, vnf_id, instance_id, mgmt_url, vnf_dict)
+            context, vnf_id, instance_id, mgmt_ip_address, vnf_dict)
         self.mgmt_create_post(context, vnf_dict)
 
         if instance_id is None or create_failed:
             return
 
-        vnf_dict['mgmt_url'] = mgmt_url
+        vnf_dict['mgmt_ip_address'] = mgmt_ip_address
 
         kwargs = {
             mgmt_constants.KEY_ACTION: mgmt_constants.ACTION_CREATE_VNF,
@@ -353,10 +377,6 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
             with excutils.save_and_reraise_exception():
                 self.delete_vnf(context, vnf_id)
 
-        if instance_id is None:
-            self._create_vnf_post(context, vnf_id, None, None,
-                                  vnf_dict)
-            return
         vnf_dict['instance_id'] = instance_id
         return vnf_dict
 
@@ -391,7 +411,7 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
                 #  key value string pairs in vnf attributes table
                 vnf_attributes['param_values'] = yaml.safe_dump(param)
             else:
-                self._report_deprecated_yaml_str()
+                raise vnfm.InvalidAPIAttributeType(atype=type(param))
         if vnf_attributes.get('config'):
             config = vnf_attributes['config']
             if isinstance(config, dict):
@@ -400,7 +420,7 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
                 #  key value string pairs in vnf attributes table
                 vnf_attributes['config'] = yaml.safe_dump(config)
             else:
-                self._report_deprecated_yaml_str()
+                raise vnfm.InvalidAPIAttributeType(atype=type(config))
 
         vnf_dict = self._create_vnf(context, vnf_info, vim_auth, infra_driver)
 
@@ -424,8 +444,8 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
         self._create_vnf_wait(context, vnf_dict, vim_auth, infra_driver)
         return vnf_dict
 
-    def _update_vnf_wait(self, context, vnf_dict, vim_auth, driver_name):
-        instance_id = self._instance_id(vnf_dict)
+    def _update_vnf_wait(self, context, vnf_dict, vim_auth, driver_name,
+                         vnf_heal=False):
         kwargs = {
             mgmt_constants.KEY_ACTION: mgmt_constants.ACTION_UPDATE_VNF,
             mgmt_constants.KEY_KWARGS: {'vnf': vnf_dict},
@@ -437,9 +457,15 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
         try:
             self._vnf_manager.invoke(
                 driver_name, 'update_wait', plugin=self,
-                context=context, vnf_id=instance_id, auth_attr=vim_auth,
+                context=context, vnf_dict=vnf_dict, auth_attr=vim_auth,
                 region_name=region_name)
             self.mgmt_call(context, vnf_dict, kwargs)
+        except vnfm.VNFUpdateWaitFailed as e:
+            with excutils.save_and_reraise_exception():
+                new_status = constants.ERROR
+                self._vnf_monitor.delete_hosting_vnf(vnf_dict['id'])
+                self.set_vnf_error_status_reason(context, vnf_dict['id'],
+                                                 six.text_type(e))
         except exceptions.MgmtDriverException as e:
             LOG.error('VNF configuration failed')
             new_status = constants.ERROR
@@ -449,8 +475,21 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
         vnf_dict['status'] = new_status
         self.mgmt_update_post(context, vnf_dict)
 
-        self._update_vnf_post(context, vnf_dict['id'],
-                              new_status, vnf_dict)
+        if vnf_heal:
+            # Update vnf status to 'ACTIVE' so that monitoring can be resumed.
+            evt_details = ("Ends the heal vnf request for VNF '%s'" %
+                           vnf_dict['id'])
+            self._vnf_monitor.update_hosting_vnf(vnf_dict, evt_details)
+            # _update_vnf_post() method updates vnf_status and mgmt_ip_address
+            self._update_vnf_post(context, vnf_dict['id'],
+                                  new_status, vnf_dict,
+                                  constants.PENDING_HEAL,
+                                  constants.RES_EVT_HEAL)
+
+        else:
+            self._update_vnf_post(context, vnf_dict['id'], new_status,
+                                  vnf_dict, constants.PENDING_UPDATE,
+                                  constants.RES_EVT_UPDATE)
 
     def update_vnf(self, context, vnf_id, vnf):
         vnf_attributes = vnf['vnf']['attributes']
@@ -462,8 +501,9 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
                 #  key value string pairs in vnf attributes table
                 vnf_attributes['config'] = yaml.safe_dump(config)
             else:
-                self._report_deprecated_yaml_str()
-        vnf_dict = self._update_vnf_pre(context, vnf_id)
+                raise vnfm.InvalidAPIAttributeType(atype=type(config))
+        vnf_dict = self._update_vnf_pre(context, vnf_id,
+                                        constants.PENDING_UPDATE)
         driver_name, vim_auth = self._get_infra_driver(context, vnf_dict)
         instance_id = self._instance_id(vnf_dict)
 
@@ -483,12 +523,48 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
                 self.mgmt_update_post(context, vnf_dict)
                 self._update_vnf_post(context, vnf_id,
                                       constants.ERROR,
-                                      vnf_dict)
+                                      vnf_dict, constants.PENDING_UPDATE,
+                                      constants.RES_EVT_UPDATE)
 
         self.spawn_n(self._update_vnf_wait, context, vnf_dict, vim_auth,
                      driver_name)
         return vnf_dict
 
+    def heal_vnf(self, context, vnf_id, heal_request_data_obj):
+        vnf_dict = self._update_vnf_pre(context, vnf_id,
+                                        constants.PENDING_HEAL)
+        driver_name, vim_auth = self._get_infra_driver(context, vnf_dict)
+        # Update vnf status to 'PENDING_HEAL' so that monitoring can
+        # be paused.
+        evt_details = ("Starts heal vnf request for VNF '%s'. "
+                       "Reason to Heal VNF: '%s'" % (vnf_dict['id'],
+                       heal_request_data_obj.cause))
+        self._vnf_monitor.update_hosting_vnf(vnf_dict, evt_details)
+
+        try:
+            self.mgmt_update_pre(context, vnf_dict)
+            self._vnf_manager.invoke(
+                driver_name, 'heal_vdu', plugin=self,
+                context=context, vnf_dict=vnf_dict,
+                heal_request_data_obj=heal_request_data_obj)
+        except vnfm.VNFHealFailed as e:
+            with excutils.save_and_reraise_exception():
+                vnf_dict['status'] = constants.ERROR
+                self._vnf_monitor.delete_hosting_vnf(vnf_id)
+                self.set_vnf_error_status_reason(context,
+                                                 vnf_dict['id'],
+                                                 six.text_type(e))
+                self.mgmt_update_post(context, vnf_dict)
+                self._update_vnf_post(context, vnf_id,
+                                      constants.ERROR,
+                                      vnf_dict, constants.PENDING_HEAL,
+                                      constants.RES_EVT_HEAL)
+
+        self.spawn_n(self._update_vnf_wait, context, vnf_dict, vim_auth,
+                     driver_name, vnf_heal=True)
+
+        return vnf_dict
+
     def _delete_vnf_wait(self, context, vnf_dict, auth_attr, driver_name):
         instance_id = self._instance_id(vnf_dict)
         e = None
@@ -595,13 +671,13 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
             return result
 
         # post
-        def _handle_vnf_scaling_post(new_status, mgmt_url=None):
+        def _handle_vnf_scaling_post(new_status, mgmt_ip_address=None):
             status = _get_status()
             result = self._update_vnf_scaling_status(context,
                                                      policy,
                                                      [status],
                                                      new_status,
-                                                     mgmt_url)
+                                                     mgmt_ip_address)
             LOG.debug("Policy %(policy)s vnf is at %(status)s",
                       {'policy': policy['name'],
                        'status': new_status})
@@ -638,7 +714,7 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
             try:
                 LOG.debug("Policy %s action is in progress",
                           policy['name'])
-                mgmt_url = self._vnf_manager.invoke(
+                mgmt_ip_address = self._vnf_manager.invoke(
                     infra_driver,
                     'scale_wait',
                     plugin=self,
@@ -650,7 +726,7 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
                 )
                 LOG.debug("Policy %s action is completed successfully",
                           policy['name'])
-                _handle_vnf_scaling_post(constants.ACTIVE, mgmt_url)
+                _handle_vnf_scaling_post(constants.ACTIVE, mgmt_ip_address)
                 # TODO(kanagaraj-manickam): Add support for config and mgmt
             except Exception as e:
                 LOG.error("Policy %s action is failed to complete",
@@ -674,14 +750,11 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
 
         return policy
 
-    def _report_deprecated_yaml_str(self):
-        utils.deprecate_warning(what='yaml as string',
-                                as_of='N', in_favor_of='yaml as dictionary')
-
     def _make_policy_dict(self, vnf, name, policy):
         p = {}
         p['type'] = policy.get('type')
-        p['properties'] = policy.get('properties') or policy.get('triggers')
+        p['properties'] = policy.get('properties') or policy.get(
+            'triggers') or policy.get('reservation')
         p['vnf'] = vnf
         p['name'] = name
         p['id'] = uuidutils.generate_uuid()
@@ -745,13 +818,26 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
         if policies:
             return policies[0]
 
-        raise exceptions.VnfPolicyTypeInvalid(type=constants.POLICY_ALARMING,
-                                              vnf_id=vnf_id)
+        raise exceptions.VnfPolicyTypeInvalid(
+            type=policy_type, policy=None,
+            valid_types=constants.VALID_POLICY_TYPES)
 
     def _validate_alarming_policy(self, context, vnf_id, trigger):
         # validate alarm status
-        if not self._vnf_alarm_monitor.process_alarm_for_vnf(vnf_id, trigger):
-            raise exceptions.AlarmUrlInvalid(vnf_id=vnf_id)
+
+        # Trigger will contain only one action in trigger['trigger'] as it
+        # filtered in _get_vnf_triggers().
+        # Getting action from trigger to decide which process_alarm_for_vnf
+        # method will be called.
+        if list(trigger['trigger'])[0]\
+                in constants.RESERVATION_POLICY_ACTIONS:
+            if not self._vnf_reservation_monitor.process_alarm_for_vnf(
+                    vnf_id, trigger):
+                raise exceptions.AlarmUrlInvalid(vnf_id=vnf_id)
+        else:
+            if not self._vnf_alarm_monitor.process_alarm_for_vnf(
+                    vnf_id, trigger):
+                raise exceptions.AlarmUrlInvalid(vnf_id=vnf_id)
 
         # validate policy action. if action is composite, split it.
         # ex: respawn%notify
@@ -789,8 +875,12 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
         # validate url
 
     def _get_vnf_triggers(self, context, vnf_id, filters=None, fields=None):
-        policy = self.get_vnf_policy_by_type(
-            context, vnf_id, policy_type=constants.POLICY_ALARMING)
+        if filters.get('name') in constants.RESERVATION_POLICY_ACTIONS:
+            policy = self.get_vnf_policy_by_type(
+                context, vnf_id, policy_type=constants.POLICY_RESERVATION)
+        else:
+            policy = self.get_vnf_policy_by_type(
+                context, vnf_id, policy_type=constants.POLICY_ALARMING)
         triggers = policy['properties']
         vnf_trigger = dict()
         for trigger_name, trigger_dict in triggers.items():
@@ -832,7 +922,7 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
                         if vnf_dict['status'] != constants.ACTIVE:
                             LOG.info(_("Scaling Policy action "
                                        "skipped due to status "
-                                       "%(status)s for vnf %(vnfid)s"),
+                                       "%(status)s for vnf %(vnfid)s") %
                                      {"status": vnf_dict['status'],
                                       "vnfid": vnf_dict['id']})
                             return
@@ -862,13 +952,16 @@ class VNFMPlugin(vnfm_db.VNFMPluginDb, VNFMMgmtMixin):
     def get_vnf_resources(self, context, vnf_id, fields=None, filters=None):
         vnf_info = self.get_vnf(context, vnf_id)
         infra_driver, vim_auth = self._get_infra_driver(context, vnf_info)
+        region_name = vnf_info.get('placement_attr', {}).\
+            get('region_name', None)
         if vnf_info['status'] == constants.ACTIVE:
             vnf_details = self._vnf_manager.invoke(infra_driver,
                                                    'get_resource_info',
                                                    plugin=self,
                                                    context=context,
                                                    vnf_info=vnf_info,
-                                                   auth_attr=vim_auth)
+                                                   auth_attr=vim_auth,
+                                                   region_name=region_name)
             resources = [{'name': name,
                           'type': info.get('type'),
                           'id': info.get('id')}
